local lib = LibStub:NewLibrary("LootEvents", "1.0")
if not lib then return nil end

--[[// Usage
	Callbacks recieve (event, ...)
	LootEvents:RegisterLootCallback(func)
		[ func(event, ...) ]
		event: 'item'
			player_name, item_link, num_items
		event: 'coin'
			player_name, total_copper, coin_string

	LootEvents:RegisterGroupCallback(func)
		Triggers func when "Group" loot mode events are recieved
		event: 'selected'
			player_name, item_link, roll_type
		event: 'rolled'
			player_name, item_link, roll_type, roll_number
		event: 'won'
			player_name, item_link
		event: 'allpass'
			item_link
--]]

-- Callback handling
lib.callbacks = lib.callbacks or { loot = {}, group = {} }
local lootcb, groupcb = lib.callbacks.loot, lib.callbacks.group

local need_loot = (#lootcb > 0)
local need_group = (#groupcb > 0)

function lib:RegisterLootCallback(func)
	table.insert(lootcb, func)
	need_loot = true
end

function lib:RegisterGroupCallback(func)
	table.insert(groupcb, func)
	need_group = true
end

local function trigger_loot(event, ...)
	for _, func in ipairs(lootcb) do
		func(event, ...)
	end
end

local function trigger_group(event, ...)
	for _, func in ipairs(groupcb) do
		func(event, ...)
	end
end


local invert_cache, smatch = {}, string.match
local function extract(str, pat)
	local inverted
	if not invert_cache[pat] then
		inverted = pat
		inverted = inverted:gsub("%%d", "(.+)")
		inverted = inverted:gsub("%%s", "(.+)")
		invert_cache[pat] = inverted
	else
		inverted = invert_cache[pat]
	end
	return smatch(str, inverted)
end

if GetCVar("showLootSpam") == "0" then
	SetCVar("showLootSpam", 1)
end

local activerolls = 0
local lootpatterns, rollhandlers, unsortedloot, currentsort, lootmethod

-- Chatmsg handler
local function Handler(text)
	if need_group then
		lootmethod = GetLootMethod()
		-- Handle group loot mode only if we're in the mode and there's a roll going
		if (lootmethod == 'group' or lootmethod == 'needbeforegreed') and activerolls > 0 then
			-- Move through the patterns one by one, match against the message
			for k, v in ipairs(rollhandlers) do
				local pat, func = v[1], v[2]
				local m1, m2, m3, m4 = extract(text, pat)
				-- Match was found, call the handler with all captured values
				if m1 then
					return func(m1, m2, m3, m4)
				end
			end
		end
	end

	if need_loot then
		-- Optimize by grouped/not and weight patterns by usage
		-- Check group mode (1 corresponds to ungrouped, 2 to grouped)
		local group = (GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) and 2 or 1
		-- If we're not currently sorting for our current mode, or it's been too long since last sort, sort ourpatterns for efficiency
		if not currentsort or currentsort ~= group or unsortedloot > 100 then
			unsortedloot, currentsort = 0, group
			table.sort(lootpatterns, function(a, b) return a[group] < b[group] end)
		else
			unsortedloot = unsortedloot + 1
		end

		-- Match string against our patterns
		for i, v in ipairs(lootpatterns) do
			local pat, func = v[3], v[4]
			local m1, m2, m3, m4 = extract(text, pat)
			-- Match found, add to counter and call pattern's handler
			if m1 then
				v[group] = v[group] + 1
				return func(m1, m2, m3, m4)
			end
		end
	end
end

-- Invert gold patterns
local function invert(pstr)
	return pstr:gsub("%%d", "(%1+)")
end
local cg = invert(GOLD_AMOUNT)
local cs = invert(SILVER_AMOUNT)
local cc = invert(COPPER_AMOUNT)

-- Parse coin strings (Really?)
local function ParseCoinString(tstr)
	local g = tstr:match(cg) or 0
	local s = tstr:match(cs) or 0
	local c = tstr:match(cc) or 0
	return g*10000+s*100+c
end

if lib.frame then
	lib.frame:SetScript("OnEvent", function() return nil end)
end
lib.frame = CreateFrame("Frame")
local f = lib.frame
local events = {}
f:SetScript("OnEvent", function(_, e, ...) if events[e] then events[e](...) end end)
local function event(e, func)
	f:RegisterEvent(e)
	events[e] = func
end

event("CHAT_MSG_LOOT", Handler)
event("CHAT_MSG_MONEY", Handler)

-- Register for the loopback, called in XLoot Monitor's options, mainly
--self:RegisterEvent("SpecialEvents_LootDummy", Handler)

-- Incriment and deincement rolls to only match while there is a roll happening
event("START_LOOT_ROLL", function() 
	activerolls = activerolls + 1
	end)
-- 2 second delay needed?
local max = math.max
event("CANCEL_LOOT_ROLL", function() 
	activerolls = max(0, activerolls - 1)
end)

local player = UnitName('player')

-- Generate loot pattern handlers
lootpatterns = { }

-- Base loot handler function. Checks argument types, adds handler to table
local lhandler = function(str, func)
	assert(type(str), "string", "First argument of lhandler() must be a string, instead, was ["..type(str).."]")
	assert(type(func) == "function" or not func, "Second argument of lhandler() must be a function, was ["..type(str).."]")
	table.insert(lootpatterns, { 0, 0, str, func })
end

-- Loot triggers
local loot = function(who, what, num) trigger_loot('item', who, what, num or 1) end
local coin = function(who, str) trigger_loot('coin', who, ParseCoinString(str), str) end

-- Add item patterns
lhandler(LOOT_ITEM, loot)
lhandler(LOOT_ITEM_MULTIPLE, loot)
lhandler(LOOT_ITEM_PUSHED_SELF, function(what) loot(player, what) end)
lhandler(LOOT_ITEM_PUSHED_SELF_MULTIPLE, function(what, num) loot(player, what, num) end)
lhandler(LOOT_ITEM_SELF, function(what) loot(player, what) end)
lhandler(LOOT_ITEM_SELF_MULTIPLE, function(what, num) loot(player, what, num) end)

-- Add coin patterns
lhandler(LOOT_MONEY, function(who, str) coin(who, str) end)
lhandler(LOOT_MONEY_SPLIT, function(str) coin(player, str) end)
lhandler(YOU_LOOT_MONEY, function(str) coin(player, str) end)

-- Generate group loot mode handlers
rollhandlers = { }

-- Base handler function. Checks argument types, adds to handler table (Presorted)
--    The pattern is stored as the first key, and the function to be called in the second
local handler = function(str, func)
	assert(type(str) == "string", "First argument of handler() must be a string, was a "..type(str))
	assert(type(func) == "function" or not func, "Second argument of handler() must be a function, and must exist")
	table.insert(rollhandlers, { str, func })
end

-- Figures the appropriate owner of the event
local ncheck = function(own, who)
	if own or who == YOU then
		return player
	end
	return who
end

-- Roll selected handler (You select Greed for XXXX, etc)
local selected = function(str, rtype, own)
	handler(str, function(who, item)
		if own then
			item, who = who, player
		end
		trigger_group('selected', who, item, rtype)
	end)
end

-- Rolled handler (XXXX roll - # by xxxxx, etc)
local rolled = function(str, rtype, own)
	handler(str, function(roll, item, who)
		trigger_group('rolled', ncheck(own, who), item, rtype, roll) 
	end)
end

-- Add handlers using construct functions, ordered specifically
selected(LOOT_ROLL_DISENCHANT, 'de')
selected(LOOT_ROLL_GREED, 'greed')
selected(LOOT_ROLL_NEED, 'need')
selected(LOOT_ROLL_DISENCHANT_SELF, 'de', true)
selected(LOOT_ROLL_GREED_SELF, 'greed', true)
selected(LOOT_ROLL_NEED_SELF, 'need', true)
selected(LOOT_ROLL_PASSED_AUTO, 'pass')
selected(LOOT_ROLL_PASSED_AUTO_FEMALE, 'pass')
selected(LOOT_ROLL_PASSED_SELF_AUTO, 'pass', true)
handler(LOOT_ROLL_ALL_PASSED, function(item) trigger_group('allpass', item) end)
selected(LOOT_ROLL_PASSED_SELF, 'pass', true)
selected(LOOT_ROLL_PASSED, 'pass')
rolled(LOOT_ROLL_ROLLED_DE, 'de')
rolled(LOOT_ROLL_ROLLED_GREED, 'greed')
rolled(LOOT_ROLL_ROLLED_NEED, 'need') 
handler(LOOT_ROLL_WON, function(who, item) trigger_group('won', ncheck(own, who), item) end)